﻿module hdk.sdlmixer;
private import sdl_mixer;
private import std.string;
private import std.cstream;
private import std.utf;
private import std.c.stdlib;

private SdlMixer mixerInstance;

extern (C) void deleteSdlMixer()
{
	delete mixerInstance;
	assert(!mixerInstance);
}

SdlMixer getMixerInstance()
{
	if (!mixerInstance)
	{
		mixerInstance = new SdlMixer();
		atexit(&deleteSdlMixer);
	}
	return mixerInstance;
}

void playBgm(dchar[] fileName) /// 一つだけ、ループ
{
	getMixerInstance().playBgm(toUTF8(fileName));
}

void playSe(dchar[] fileName) /// いくつでも、ループしない
{
	getMixerInstance().playSe(toUTF8(fileName));
}

void playVoice(dchar[] fileName) /// 一つだけ、ループしない
{
	getMixerInstance().playVoice(toUTF8(fileName));
}

void stop() /// 再生中のサウンドを全て停止して、サウンドリソースを全て開放する
{
	getMixerInstance().stop();
}

private class SdlMixer
{
	private import sdl_mixer;
	private import hdk.error;
	mixin SdlError;
	
	private static bit initialized = false;
	
	this()
	{
		if (initialized) error(__LINE__, "既に初期化されてます");
		Mix_OpenAudio(44100, AUDIO_S16, 2, 4096);
		throwIfSdlError(__LINE__);
		initialized = true;
	}
	
	~this()
	{
		delete bgm;
		delete voice;
		delete seMap;
		Mix_CloseAudio();
		initialized = false;
	}
	
	private Bgm bgm;
	private Voice voice;
	private SeMap seMap;
	
	void playBgm(char[] fileName)
	{
		delete bgm;
		bgm = new Bgm(fileName);
		throwIfSdlError(__LINE__);
	}
	
	void playVoice(char[] fileName)
	{
		delete voice;
		voice = new Voice(fileName);
		throwIfSdlError(__LINE__);
	}
	
	void playSe(char[] fileName)
	{
		if (!seMap) seMap = new SeMap();
		if (!seMap.has(fileName)) seMap.add(fileName);
		seMap.play(fileName);
		throwIfSdlError(__LINE__);
	}
	
	void stop()
	{
		delete bgm;
		delete voice;
		delete seMap;
		throwIfSdlError(__LINE__);
	}
}

private class Bgm
{
	private Mix_Music* bgm = null;
	
	this(char[] fileName)
	{
		bgm = Mix_LoadMUS(toStringz(fileName));
		Mix_PlayMusic(bgm, -1);
	}
	
	~this()
	{
		Mix_HaltMusic();
		Mix_FreeMusic(bgm);
	}
}

private class Chunk
{
	protected Mix_Chunk* chunk;
	
	private static Mix_Chunk* createChunk(char[] fileName)
	{
		return Mix_LoadWAV_RW(SDL_RWFromFile(toStringz(fileName), "rb"), 1);
	}
}

private class Voice : Chunk
{
	private const uint VOICE_CHANNEL = MIX_CHANNELS - 1;
	
	this(char[] fileName)
	{
		chunk = createChunk(fileName);
		Mix_PlayChannelTimed(VOICE_CHANNEL, chunk, 0, -1);
	}
	
	~this()
	{
		Mix_HaltChannel(VOICE_CHANNEL);
		Mix_FreeChunk(chunk);
	}
}

private class Se : Chunk
{
	this(char[] fileName)
	{
		chunk = createChunk(fileName);
	}
	
	~this()
	{
		Mix_FreeChunk(chunk);
	}
	
	void play()
	{
		Mix_PlayChannelTimed(-1, chunk, 0, -1);
	}
}

private class SeMap
{
	private Se[char[]] seMap;
	
	~this()
	{
		stopSe();
		foreach (Se a; seMap.values) delete a;
	}
	
	bit has(char[] key) { return key in seMap ? true : false; }
	void add(char[] key) { seMap[key] = new Se(key); }
	void play(char[] key) { seMap[key].play(); }
	
	private static void stopSe()
	{
		const uint SE_CHANNEL_LENGTH = MIX_CHANNELS - 1;
		for (int i = 0; i < SE_CHANNEL_LENGTH; i++) Mix_HaltChannel(i);
	}
}
