﻿module nanami.model;
private import hdk.window;
private import std.file;
private import std.utf;
private import hdk.zipex;
private import std.string;
private import nanami.skeleton;
private import nanami.blackboard;
private import std.cstream;

interface _Model
{
	//this(dchar[] nanamiFileName)
	void alive(bit a);
	void modelView(Matrix mv);
	void projection(Matrix p);
	void lightDirection(Vector a);
	Subset opIndex(uint y, uint x);
	uint width(uint yIndex);
	uint height();
	MotionIterator motionIterator();
	bit hasMotion();
	uint motionLength();
	bit hasSkeleton();
	uint boneLength();
	Bone getBone(uint index);
	void setBone(uint index, Bone bone); // boneを一時的に置き換える
	void lookAt(uint boneIndex, Vector at, Vector up);
}

///
interface MotionIterator
{
	uint x(); /// 現在のモーションのフレーム位置
	void x(uint a); ///
	uint y(); /// モーションのインデックス
	void y(uint a); ///
	bit isRangeOut(); ///
	uint width(); /// 現在のモーションのフレーム数
	uint height(); /// モーションの数
	void opPostInc(); /// x++
 	void opPostDec(); /// x--
}

///
interface Subset
{
	bit visible(); ///
	void visible(bit a); ///
	bit exists(); ///
}

///
class Model : _Model, TerminableCommand
{
	private dchar[] TOON_SHADER = "toon";
	private dchar[] TOON_BASE = "toonbase.png";
	private dchar[] SMESH_TOON_SHADER = "smesh_toon";
	private final dchar[] nanamiFileName;
	private bit[][] visible;
	private Skeleton skeleton;
	private MotionMatrix[] motionMatrix;
	
	this(dchar[] nanamiFileName, bit trace = false) ///
	{
		if (!exists(toUTF8(nanamiFileName)))
		{
			throw new Error("file name?" ~ toUTF8(nanamiFileName));
		}
		this.nanamiFileName = nanamiFileName;
		
		uint yLength = 0;
		while (true)
		{
			if (!BufferedZipFile.isZipEntry(vertexBufferName(0, yLength))) break;
			yLength++;
		}
		visible.length = yLength;
		if (trace) { dout.writefln("yLength=%d", yLength); dout.flush(); }
		
		for (uint y = 0; y < visible.length; y++)
		{
			uint xLength = 0;
			while (true)
			{
				if (!BufferedZipFile.isZipEntry(vertexBufferName(xLength, y))) break;
				xLength++;
			}
			visible[y].length = xLength;
			if (trace) { dout.writefln("y=%d xLength=%d", y, xLength); dout.flush(); }
		}
		
		for (int i = 0; i < visible[0].length; i++) visible[0][i] = true;
		
		_modelView = _projection = Matrix.identity;
		_lightDirection = Vector.create(0, -1, 0);
		
		if (BufferedZipFile.isZipEntry(nanamiFileName ~ `\skeleton_bones`))
		{
			ZipArchiveEx decoder = BufferedZipFile.getZipArchiveEx(nanamiFileName);
			skeleton = new Skeleton(
				cast(float[])decoder.get("skeleton_bones"),
				cast(ubyte[])decoder.get("skeleton_relations")
				);
			if (trace) { dout.writeLine("loaded skeleton_bones"); dout.flush(); }
		}
		else
		{
			if (trace) { dout.writeLine("not found skeleton_bones"); dout.flush(); }
		}
		
		foreach (char[] a; motionMatrixNames)
		{
			if (trace) { dout.writeString("load " ~ a ~ "..."); dout.flush(); }
			ZipArchiveEx decoder = BufferedZipFile.getZipArchiveEx(nanamiFileName);
			motionMatrix ~= new MotionMatrix(decoder.get(a), true);
			if (trace) { dout.writeLine("loaded " ~ a); dout.flush(); }
		}
		
		mi = hasMotion ? new MotionIterator_(this) : null;
		
		_modelView = Matrix.lookAt(0, 0, 200,  0, 0, 0);
		_projection = Matrix.perspectiveFov;
	}
	
	private bit _alive = true;
	bit alive() { return _alive; } ///
	void alive(bit a) { _alive = a; } ///
	
	void opCall(WindowResource wr)
	{
		wr.depthTestWriting(true, true);
		wr.cullingEnableFront(true, true);
		wr.modelView = _modelView;
		wr.projection = _projection;
		isPcnt ? drawPcnt(wr) : drawPcntmw(wr);
	}
	
	private void drawPcnt(WindowResource wr)
	{
		wr.setShaderVector(TOON_SHADER, "lightDirection", _lightDirection);
		wr.setShaderInt(TOON_SHADER, "texture0", 0);
		wr.setShaderInt(TOON_SHADER, "texture1", 1);
		foreach (dchar[] vb, dchar[] texture; this)
		{
			wr.drawPcnt(TOON_SHADER, vb, texture, TOON_BASE);
		}
	}
	
	private void drawPcntmw(WindowResource wr)
	{
		Matrix[] ma = getMatrixArray(skeleton, motionMatrix[mi.y], mi.x);
		
		if (tbmX != mi.x || tbmY != mi.y)
		{
			tempBoneMatrix = null;
			tbmX = mi.x;
			tbmY = mi.y;
		}
		foreach (uint a; tempBoneMatrix.keys) ma[a] = tempBoneMatrix[a];
		
		wr.setShaderMatrixArray(SMESH_TOON_SHADER, "matrixArray", ma);
		wr.setShaderVector(SMESH_TOON_SHADER, "lightDirection", _lightDirection);
		wr.setShaderInt(SMESH_TOON_SHADER, "texture0", 0);
		wr.setShaderInt(SMESH_TOON_SHADER, "texture1", 1);
		foreach (dchar[] vb, dchar[] texture; this)
		{
			wr.drawPcntmw(SMESH_TOON_SHADER, "matrixIndex", "weight", vb, texture, TOON_BASE);
		}
	}
	
	private bit isPcnt()
	{
		return !BufferedZipFile.isZipEntry(
			toUTF32(format(`%s\motion_matrix[0]`, nanamiFileName))
			) ? true : false;
	}
	
	private Matrix _modelView, _projection;
	void modelView(Matrix mv) { this._modelView = mv; } ///
	void projection(Matrix p) { this._projection = p; } ///
	
	private Vector _lightDirection;
	void lightDirection(Vector a) { _lightDirection = a; } ///
	
	private int opApply(int delegate(inout dchar[] vb, inout dchar[] texture) dg)
	{
		for (uint y = 0; y < visible.length; y++)
		{
			for (uint x = 0; x < visible[y].length; x++)
			{
				if (!visible[y][x]) continue;
				dchar[] vb = vertexBufferName(x, y);
				dchar[] texture = textureName(x, y);
				if (BufferedZipFile.getEntrySize(vb) <= uint.sizeof * 2) continue;
				int result = dg(vb, texture);
				if (result != 0) return result;
			}
		}
		return 0;
	}
	
	private dchar[] vertexBufferName(uint x, uint y)
	{
		return toUTF32(format(`%s\vertex_buffer[%d][%d]`, nanamiFileName, y, x));
	}
	
	private dchar[] textureName(uint x, uint y)
	{
		return toUTF32(format(`%s\texture[%d][%d]`, nanamiFileName, y, x));
	}
	
	Subset opIndex(uint y, uint x) ///
	{
		return new Subset_(this, x, y);
	}
	
	uint width(uint yIndex) { return visible[yIndex].length; } ///
	uint height() { return visible.length; } ///
	
	private char[][] motionMatrixNames()
	{
		char[][] result;
		int i = 0;
		while (true)
		{
			char[] name = format("motion_matrix[%d]", i);
			if (!BufferedZipFile.isZipEntry(nanamiFileName ~ `\` ~ toUTF32(name))) break;
			result ~= name;
			i++;
		}
		return result;
	}
	
	private static Matrix[] getMatrixArray(Skeleton skeleton, MotionMatrix motionMatrix, uint t)
	{
		Matrix[] result = new Matrix[motionMatrix.boneLength];
		for (int i = 0; i < result.length; i++) result[i] = getBoneMatrix(skeleton, motionMatrix, t, i);
		return result;
	}
	
	private static Matrix getBoneMatrix(Skeleton skeleton, MotionMatrix motionMatrix, uint t, uint boneIndex)
	{
		Matrix result = motionMatrix.getLookAt(t, boneIndex);
		uint currentIndex = boneIndex;
		while (true)
		{
			result *= motionMatrix.getRotation(t, currentIndex);
			result *= motionMatrix.getTranslation(t, currentIndex);
			if (skeleton.isRoot(currentIndex)) break;
			currentIndex = skeleton.getParent(currentIndex);
		}
		return result;
	}
	
	private MotionIterator mi;
	MotionIterator motionIterator() { return mi; } ///
	bit hasMotion() { return motionMatrix ? true : false; } ///
	uint motionLength() { return motionMatrix ? motionMatrix.length : 0; } ///
	
	bit hasSkeleton() { return skeleton ? true : false; } ///
	uint boneLength() { return hasSkeleton ? skeleton.boneList.length : 0; } ///
	Bone getBone(uint index) ///
	{
		return mul3(
			skeleton.boneList[index], getBoneMatrix(skeleton, motionMatrix[mi.y], mi.x, index)
			);
	}
	private Matrix[uint] tempBoneMatrix;
	private uint tbmX, tbmY;
	void setBone(uint index, Bone bone) ///
	{
		if (tbmX != mi.x || tbmY != mi.y)
		{
			tempBoneMatrix = null;
			tbmX = mi.x;
			tbmY = mi.y;
		}
		
		Bone newBone = bone;
		Bone oldBone = skeleton.boneList[index];
		
		Bone oldBaseBone = nanami.model.mul3(oldBone,
			inverse(Matrix.translation(
				oldBone.startPosition.x, oldBone.startPosition.y, oldBone.startPosition.z))
			* Matrix.lookAtLH(
				Vector.create(0, 0, 0),
				oldBone.endPosition - oldBone.startPosition,
				oldBone.hPosition - oldBone.startPosition)
			);
		
		Bone newBaseBone = new Bone(0, 0, 0,
			0, 0, distance(newBone.endPosition, newBone.startPosition),
			0, distance(newBone.hPosition, newBone.startPosition), 0);
		
		tempBoneMatrix[index] =
			inverse(Matrix.translation(
				oldBone.startPosition.x, oldBone.startPosition.y, oldBone.startPosition.z))
			* Matrix.lookAtLH(
				Vector.create(0, 0, 0),
				oldBone.endPosition - oldBone.startPosition,
				oldBone.hPosition - oldBone.startPosition)
			
			* Matrix.scaling(1.0,
				newBaseBone.hPosition.y / oldBaseBone.hPosition.y,
				newBaseBone.endPosition.z / oldBaseBone.endPosition.z)
			// ここまででbaseBoneに変換されている
			
			* inverse(Matrix.lookAtLH(
				Vector.create(0, 0, 0),
				newBone.endPosition - newBone.startPosition,
				newBone.hPosition - newBone.startPosition))
			* Matrix.translation(
				newBone.startPosition.x, newBone.startPosition.y, newBone.startPosition.z);
	}
	
	void lookAt(uint boneIndex, Vector at, Vector up) ///
	{
		Bone baseBone = getBone(boneIndex);
		
		Vector atDirection = normalize(at - baseBone.startPosition);
		Vector upDirection = normalize(up - baseBone.startPosition);
		
		Vector endPosition = baseBone.startPosition
			+ atDirection * distance(baseBone.endPosition, baseBone.startPosition);
		
		Vector cau = normalize(cross(atDirection, upDirection));
		Vector upDirection2 = normalize(cross(cau, atDirection));
		Vector heightPosition = baseBone.startPosition
			+ upDirection2 * distance(baseBone.hPosition, baseBone.startPosition);
		
		assert(
			distance(baseBone.startPosition, endPosition) > distance(baseBone.startPosition, heightPosition)
			);
		assert(
			distance(endPosition, heightPosition) > distance(baseBone.startPosition, endPosition)
			);
		setBone(boneIndex, new Bone(baseBone.startPosition, endPosition, heightPosition));
	}
}

Bone mul3(Bone a, Matrix b)
{
	return new Bone(
		hdk.math.mul3(a.startPosition, b),
		hdk.math.mul3(a.endPosition, b),
		hdk.math.mul3(a.hPosition, b)
		);
}

class MotionIterator_ : MotionIterator
{
	private final Model model;
	this(Model model) { this.model = model; }
	private uint _x = 0, _y;
	uint x() { return _x; }
	void x(uint a) { _x = a; }
	uint y() { return _y; }
	void y(uint a) { _y = a; }
	bit isRangeOut() { return _y >= height || _x >= width ? true : false; }
	uint width() { return model.motionMatrix[y].frameLength; }
	uint height() { return model.motionMatrix.length; }
	void opPostInc() { _x++; }
 	void opPostDec() { _x--; }
}

class Subset_ : Subset
{
	private Model model;
	private final uint x, y;
	
	this(Model model, uint x, uint y)
	{
		this.model = model;
		this.x = x;
		this.y = y;
	}
	
	bit visible()
	{
		throwIfNullError();
		return model.visible[y][x];
	}
	
	void visible(bit a)
	{
		throwIfNullError();
		model.visible[y][x] = a;
	}
	
	private void throwIfNullError()
	{
		if (!exists()) throw new Error(format("Subset[%d][%d] is null", y, x));
	}
	
	bit exists()
	{
		return y < model.visible.length && x < model.visible[y].length ? true : false;
	}
}
